#!/usr/bin/env python3

# by TheTechromancer

import itertools
from .mutator import Mutator


class Perm(Mutator):
    '''
    permutates words from iterable
    takes:      iterable containing words
    yields:     word permutations ('pass', 'word' --> 'password', 'wordpass', etc.)
    '''

    fname = 'perm'

    def __init__(self, _input, perm_depth=0, double=False):

        self.perm_depth = perm_depth
        self.double = double
        self.input = _input

        super().__init__(_input, limit=None)


    def __len__(self):

        length = len(self.input)

        if self.perm_depth > 1:
            initial_length = len(self.input)
            for i in range(2, self.perm_depth+1):
                length += initial_length ** i

        elif self.double:
            length *= 2

        # prevent division by zero
        return max(1, length)


    def __iter__(self):

        if self.perm_depth > 1:
            for d in range(1, self.perm_depth+1):
                for p in itertools.product(self.input, repeat=d):
                    yield b''.join(p)

        else:
            for word in self.input:
                yield word
                if self.double:
                    yield word + word


#!/usr/bin/env python3

# by TheTechromancer

from .mutator import Mutator


class Leet(Mutator):

    scale = 1
    fname = 'leet'

    def __init__(self, _input, limit=128):

        super().__init__(_input, limit)

        # "leet" character swaps - modify as needed.
        # Keys are replaceable characters; values are their leet replacements
        self.leet_common = self._dict_str_to_bytes({
            'a': ['@'],
            'A': ['@'],
            'e': ['3'],
            'E': ['3'],
            'i': ['1'],
            'I': ['1'],
            'o': ['0'],
            'O': ['0'],
            's': ['5', '$'],
            'S': ['5', '$'],
            't': ['7'],
            'T': ['7']
        })

        self.leet_all = self._dict_str_to_bytes({
            'a': ['4', '@'],
            'A': ['4', '@'],
            'b': ['8'],
            'B': ['8'],
            'e': ['3'],
            'E': ['3'],
            'i': ['1'],
            'I': ['1'],
            'l': ['1'],
            'L': ['1'],
            'o': ['0'],
            'O': ['0'],
            's': ['5', '$'],
            'S': ['5', '$'],
            't': ['7'],
            'T': ['7']
        })


    def mutate(self, word):

        for r in self._leet(word, swap_values=self.leet_common):
            yield r



    def _leet(self, word, swap_values=None, end=0):

        if not swap_values:
            swap_values = self.leet_all

        if len(word) == 1:
            yield word
            try:
                for leet_char in swap_values[word]:
                    yield leet_char
            except KeyError:
                pass

        else:
            mid_point = int(len(word)/2)
            if end == 0:
                for right_half in self._leet(word[mid_point:], swap_values=swap_values, end=end^1):
                    for left_half in self._leet(word[:mid_point], swap_values=swap_values, end=end^1):
                        yield left_half + right_half
            else:
                for left_half in self._leet(word[:mid_point], swap_values=swap_values, end=end^1):
                    for right_half in self._leet(word[mid_point:], swap_values=swap_values, end=end^1):
                        yield left_half + right_half



    def _dict_str_to_bytes(self, d):

        new_dict = {}
        for key in d:
            new_dict[key.encode('utf-8')] = [v.encode('utf-8') for v in d[key]]
        return new_dict


#!/usr/bin/env python3

# by TheTechromancer

# This was adapted from the Password Statistical Analysis tool by Peter Kacherginsky
# https://github.com/iphelix/pack

import re
import sys
import argparse
import webbrowser
from pathlib import Path
from password_stretcher.lib.utils import *
from password_stretcher.lib.errors import *
from password_stretcher.lib.spider import Spider
from password_stretcher.lib.stat.report import *
from password_stretcher.lib.policy import PasswordPolicy


def main():

    parser = argparse.ArgumentParser(description='PASSWORDS NED, IN AN OPEN FIELD')
    parser.add_argument('-i', '--input', nargs='+', type=Path, default=ReadSTDIN(binary=False), help='password list(s) to analyze (default: STDIN)', metavar='')
    parser.add_argument('-o', '--output', type=Path, help='save all data to this directory (it will be created)')
    reporting = argparse.ArgumentParser.add_argument_group(parser, 'report options')
    reporting.add_argument('--limit', type=int, default=20, metavar=20, help='limit the number of results per chart (default: 20)')
    reporting.add_argument('--hiderare', type=float, default=5.0, metavar='1.0', help='hide statistics covering less than this percent of the max value (default: 5.0)')
    reporting.add_argument('--title', default='Password Analysis', help='title of report')
    reporting.add_argument('--theme', choices=['light', 'dark'], default='dark', help='visual theme for report (default: dark)')
    filters = argparse.ArgumentParser.add_argument_group(parser, 'password complexity filters')
    filters.add_argument('--minlength', type=int, metavar='8', help='minimum password length')
    filters.add_argument('--maxlength', type=int, metavar='8', help='maximum password length')
    filters.add_argument('--mincharsets', type=int, metavar='3', help='must have this many character sets')
    filters.add_argument('--charsets', nargs='+', choices=PasswordPolicy.charset_choices, help='must include these character sets')
    filters.add_argument('--regex', type=re.compile, metavar='\'$[a-z]*^\'', help='custom regex')
    parsing = argparse.ArgumentParser.add_argument_group(parser, 'password file parsing options')
    parsing.add_argument('-d', '--delimiter', metavar=':', help='file delimiter')
    parsing.add_argument('-f', '--field', type=int, default=2, metavar='2', help='password field number (default: 2)')
    webserver = argparse.ArgumentParser.add_argument_group(parser, 'web server options options')
    webserver.add_argument('--port', type=int, default=8050, metavar='80', help='port to listen on (default: 8050)')
    webserver.add_argument('--host', default='127.0.0.1', metavar='0.0.0.0', help='interface to listen on (default: 127.0.0.1)')
    webserver.add_argument('--no-browser', action='store_true', help='don\'t attempt to open a web browser')

    try:

        options = parser.parse_args()

        # print help if there's nothing to analyze
        if type(options.input) == ReadSTDIN and stdin.isatty():
            parser.print_help()
            sys.stderr.write('\n\n[!] Please specify password list(s) or pipe to STDIN\n')
            exit(2)

        if not type(options.input) == ReadSTDIN:
            print(f'[*] Analyzing passwords in {", ".join([str(f) for f in options.input])}')
            options.input = ReadFiles(*options.input, binary=False)

        if type(options.input) == Spider:
            options.input.depth = options.spider_depth
            options.input.start()

        report = PasswordReport(options)
        report.stats.analyze(options.input)

        if options.output:
            report.dump_everything(options.output)

        report.make_html_report()


    except re.error:
        print(f'[!] Invalid regex')
        print('[*] Remember to place regex in single quotes: \'^Password[0-9]+\'')
        sys.exit(2)
    except BrokenPipeError:
        # Python flushes standard streams on exit; redirect remaining output
        # to devnull to avoid another BrokenPipeError at shutdown
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        sys.exit(1)  # Python exits with error code 1 on EPIPE
    except (PasswordStretcherError, AssertionError) as e:
        sys.stderr.write(f'\n[!] {e}\n')
        exit(1)
    except argparse.ArgumentError:
        sys.stderr.write('\n[!] Check your syntax. Use -h for help.\n')
        exit(2)
    except KeyboardInterrupt:
        sys.stderr.write('\n[!] Interrupted.\n')
        exit(2)


if __name__ == '__main__':
    main()


# by TheTechromancer

import re
import numpy as np
import pandas as pd
from math import log, e as euler
from collections import OrderedDict
from password_stretcher.lib.policy import PasswordPolicy

# graphing
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots

'''
class DataFrame(pd.DataFrame):

    def custom_truncate(self, size, percent=None):

        if percent is None:
            percent = 1

        hiderare_threshold = size * (percent / 100)
        rare = self[(self[1] < hiderare_threshold) | (self.index >= self.options.limit)]
        self.drop(rare.index, inplace=True)
        return rare
'''

class Stat:

    # not unicode-friendly
    word_regex = re.compile(r'[a-z][a-z01357$@]+[a-z]', re.I)

    # sort before truncation
    presort = {'by': 1, 'ascending': False}
    # sort after truncation
    postsort = {'by': 1, 'ascending': False}

    hiderare = True
    hidesingles = True
    include_other = True

    def __init__(self, options, title=None, key_label=None, val_label=None):

        self.title = ('' if title is None else str(title))
        self.key_label = ('' if key_label is None else str(key_label))
        self.val_label = ('' if val_label is None else str(val_label))

        self.options = options
        self._json = dict()

        # caching
        self._df = None


    @property
    def json(self):

        return self._json


    def increment(self, k):
        '''
        override if needed
        '''

        self._increment(k)


    def make_figure(self, theme):

        return self.make_bar(title=self.title, theme=theme)


    def make_bar(self, title=None, theme=None):

        #fig = px.bar(df, title=title, x=df.columns[0], y=df.columns[1], template=theme)
        fig = go.Figure(data=[self._make_bar()])
        fig.update_layout(
            title_text=title,
            template=theme,
            font=dict(size=20),
            legend={
                'bordercolor': 'rgba(255, 255, 255, 0)'
            }
        )

        fig.update_xaxes(type='category')
        return fig


    def _make_bar(self, name=None):

        if name is None:
            name = self.title

        return go.Bar(x=self.df.iloc[:, 0], y=self.df.iloc[:, 1], name=name)


    def _make_scatter(self, name=None):

        if name is None:
            name = self.title

        return go.Scatter(x=self.df.iloc[:, 0], y=self.df.iloc[:, 1], name=name, mode='markers')


    @staticmethod
    def make_pie(df, title=None, theme=None):

        #fig = px.pie(df, title=title, names=df.columns[0], values=df.columns[1])
        fig = go.Figure(data=[go.Pie(labels=df.iloc[:, 0], values=df.iloc[:, 1], hole=.3)])
        fig.update_layout(title_text=title, height=650, template=theme, font=dict(size=20))
        return fig


    def _increment(self, k):

        try:
            self._json[k] += 1
        except KeyError:
            self._json[k] = 1


    @property
    def df(self):
        '''
        returns self as pandas data frame
        '''

        #if self._df is None:

        # convert first column to strings for sorting
        #df = pd.DataFrame(data).astype({0: str})
        df = pd.DataFrame(self.json.items())
        if not df.empty:
            # sort before dropping takes place
            df.sort_values(**self.presort, ignore_index=True, inplace=True)

            # combine rare and truncated data points
            rare = (
                df[
                    (df[1] < df[1].max() * (self.options.hiderare / 100) if self.hiderare else df[1] == '_eyes_') | \
                    (df[1] == 1 if self.hidesingles else df[1] == '_throat_') | \
                    (df.index >= self.options.limit if True else df[1] == '_genitals_')
                ]
            )

            '''
            if self.hiderare:
                hiderare_threshold = df[1].max() * (self.hiderare / 100)
                rare = df[(df[1] < hiderare_threshold) | (df.index >= self.options.limit)]
            else:
                rare = df[df.index >= self.options.limit]

            if self.hidesingles:
                hiddensingles = df[df[1] == 1]
                rare = pd.concat([rare, hiddensingles])
            '''

            df.drop(rare.index, inplace=True)

            # sort after dropping has taken place
            df.sort_values(**self.postsort, ignore_index=True, inplace=True)

            # check if there's already an 'Other' category
            if self.include_other:
                other = df[df[0]=='Other']
                if other.empty:
                    df = df.append(pd.DataFrame([('Other', rare[1].sum())]), ignore_index=True)
                # if so, add to it instead
                else:
                    df.loc[df[0]=='Other',1] += rare[1].sum()

            # rename labels
            df.rename(columns={0: self.key_label, 1: self.val_label}, inplace=True)

        self._df = df

        return self._df





class CharacterSetStat(Stat):

    @property
    def json(self):

        return {PasswordPolicy.charset_flags[k]: v for k,v in self._json.items()}


class PasswordStat(Stat):

    hiderare = False
    include_other = False


class MutationStat(Stat):

    include_other = False

    def increment(self, password):

        for match in self.word_regex.finditer(password):
            mutation = f'{password[:match.span()[0]]}password{password[match.span()[1]:]}'
            super().increment(mutation)


class PasswordLengthStat(Stat):

    postsort = {'by': 0}



class PasswordEntropyStat(Stat):

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self._json['coordinates'] = dict()


    def increment(self, password):

        try:
            self._json['coordinates'][(self.entropy(password), len(password))] += 1
        except KeyError:
            self._json['coordinates'][(self.entropy(password), len(password))] = 1


    def _make_scatter(self, *args, name=None, **kwargs):

        if name is None:
            name = self.title

        size = self.df.iloc[:, 2]
        fig = go.Scatter(
            x=self.df.iloc[:, 0],
            y=self.df.iloc[:, 1],
            mode='markers',
            marker={
                'size': size,
                'sizemode': 'area',
                'sizeref': 2.*max(size)/(40.**2),
                'sizemin': 3
            },
            name=name
        )

        return fig


    @property
    def df(self):

        df = pd.DataFrame([(a, b, c) for ((a,b) ,c) in self._json['coordinates'].items()])
        df.rename(columns={0: self.key_label, 1: self.val_label, 2: 'Count'}, inplace=True)
        return df


    @staticmethod
    def entropy(password, base=None):

        password = list(password)
        n_labels = len(password)

        if n_labels <= 1:
            return 0

        value,counts = np.unique(password, return_counts=True)
        probs = counts / n_labels
        n_classes = np.count_nonzero(probs)

        if n_classes <= 1:
            return 0

        ent = 0.

        # Compute entropy
        base = euler if base is None else base
        for i in probs:
            ent -= i * log(i, base)

        return ent



class PasswordMetaStat(Stat):

    def __init__(self, cracked, uncracked, complex_count, noncomplex_count, complex_length, noncomplex_length, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.cracked = cracked
        self.uncracked = uncracked
        self.complex_count = complex_count
        self.noncomplex_count = noncomplex_count
        self.complex_length = complex_length
        self.noncomplex_length = noncomplex_length
        self.total = cracked + uncracked


    def make_figure(self, *args, **kwargs):

        return self.make_sunburst(*args, **kwargs)


    def make_sankey(self, theme):

        fig = go.Figure(data=[go.Sankey(
            valueformat = ".0f",
            valuesuffix = "TWh",
            # Define nodes
            node = dict(
              pad = 15,
              thickness = 15,
              line = dict(color = "black", width = 0.5),
              label =  data['data'][0]['node']['label'],
              color =  data['data'][0]['node']['color']
            ),
            # Add links
            link = dict(
              source =  data['data'][0]['link']['source'],
              target =  data['data'][0]['link']['target'],
              value =  data['data'][0]['link']['value'],
              label =  data['data'][0]['link']['label'],
              color =  data['data'][0]['link']['color']
        ))])
        return fig


    def make_sunburst(self, theme):

        root = f'{self.total:,}'

        labels = [root]
        values = ['']
        parents = ['']

        # cracked layer
        if self.uncracked:
            labels += ['Cracked', 'Uncracked']
            values += [self.cracked, self.uncracked]
            parents += [root] * 2

        if self.noncomplex_count > 0:
            # complexity layer
            labels += ['Compliant', 'Non-Compliant']
            values += [self.complex_count, self.noncomplex_count]
            if self.uncracked:
                parents += [f'Cracked'] * 2
            else:
                parents += [root] * 2

        # length layers
        '''
        labels += [str(_) for _ in self.complex_length.df.iloc[:, 0]]
        values += list(self.complex_length.df.iloc[:, 1])
        parents += ['Meets Complexity'] * len(self.complex_length.df.iloc[:, 1])

        labels += [str(_) for _ in self.noncomplex_length.df.iloc[:, 0]]
        values += list(self.noncomplex_length.df.iloc[:, 1])
        parents += ['Fails Complexity'] * len(self.noncomplex_length.df.iloc[:, 1])
        '''

        fig =px.sunburst(
            {
                'labels': labels,
                'values': values,
                'parents': parents,
            },
            names='labels',
            values='values',
            parents='parents',
            branchvalues='total',
            color='labels',
            color_discrete_sequence=px.colors.qualitative.Plotly,
            color_discrete_map={
                'Cracked': px.colors.qualitative.Plotly[1],
                'Uncracked': px.colors.qualitative.Plotly[0],
                'Compliant': px.colors.qualitative.Plotly[0],
                'Non-Compliant': px.colors.qualitative.Plotly[1],
            },
        )

        fig.update_layout(
            height=600,
            template=theme,
            margin=dict(t=0, l=0, r=0, b=0),
            font_color='white',
            font=dict(size=20),
            title_font_color='white',
            paper_bgcolor='rgba(0,0,0,0)',
            plot_bgcolor='rgba(0,0,0,0)'
        )


        '''
        fig =go.Figure(go.Sunburst(
            labels=labels,
            parents=parents,
            values=values,
            branchvalues='total',
            sort=False,
            #color_discrete_map={'Cracked': 'red', 'Uncracked': 'green'}
        ))

        #fig.update_xaxes(tickfont=dict(color='rgba(0,0,0,0)'))
        #fig.update_yaxes(tickfont=dict(color='rgba(0,0,0,0)'))
        '''

        return fig



class AdvancedMaskStat(Stat):

    def make_figure(self, theme):

        import plotly.figure_factory as ff

        fig = make_subplots(
            rows=1, cols=2,
            specs=[
                [{'type': 'table'}, {'type': 'pie'}]
            ]
        )

        fig.add_trace(
            go.Table(
                header={
                    'values': list(self.df.columns),
                },
                cells={
                    'values': [self.df[k].tolist() for k in self.df.columns]
                }
            ),
            row=1, col=1
        )

        fig.add_trace(
            go.Pie(
                labels=self.df.iloc[:, 0],
                values=self.df.iloc[:, 1],
                hole=.3
            ),
            row=1, col=2
        )

        fig.update_layout(title_text=self.title, height=650, template=theme)

        return fig




class BaseWordStat(Stat):

    # this should work but it doesn't
    # re.compile(r'[^\d\W]([01357@$]|[^\d\W])+[^\d\W]').findall('Pa$$wórd123')
    hiderare = False
    include_other = False

    def increment(self, k):

        words = []
        for word in self.word_regex.findall(k):
            word = word.lower()
            if not word in words:
                words.append(word)
        for word in words:
            self._increment(word)




class PasswordStats:

    def __init__(self, options):

        self.options = options

        # complexity filters
        self.policy = PasswordPolicy(
            minlength=options.minlength,
            maxlength=options.maxlength,
            mincharsets=options.mincharsets,
            required_charsets=options.charsets,
            regex=options.regex,
        )

        # global password stats
        self._meta = None

        self.overall = OrderedDict([
            ('passwords', PasswordStat(self.options, title='Top Passwords', key_label='Password', val_label='Count')),
            ('basewords', BaseWordStat(self.options, title='Top Base Words', key_label='Base Word', val_label='Count')),
            ('mutations', MutationStat(self.options, title='Top Mutations', key_label='Mutation', val_label='Count')),
            ('advancedmasks', AdvancedMaskStat(self.options, title='Advanced Masks', key_label='Mask', val_label='Count')),
        ])

        # passwords that meet complexity filters
        self.complex_counter = 0
        self.complex = OrderedDict([
            ('length', PasswordLengthStat(self.options, title='Password Length', key_label='Password Length', val_label='Count')),
            ('entropy', PasswordEntropyStat(self.options, title='Password Entropy', key_label='Entropy', val_label='Password Length')),
            ('simplemasks', Stat(self.options, title='Simple Masks', key_label='Mask', val_label='Count')),
            ('charactersets', CharacterSetStat(self.options, title='Character Sets', key_label='Character Set', val_label='Count')),
        ])

        # passwords that don't meet complexity filters
        self.noncomplex_counter = 0
        self.noncomplex = OrderedDict([
            ('length', PasswordLengthStat(self.options, title='Password Length', key_label='Password Length', val_label='Count')),
            ('entropy', PasswordEntropyStat(self.options, title='Password Entropy', key_label='Password Entropy', val_label='Password Length')),
            ('simplemasks', Stat(self.options, title='Simple Masks', key_label='Mask', val_label='Count')),
            ('charactersets', CharacterSetStat(self.options, title='Character Sets', key_label='Character Set', val_label='Count')),
        ])

        self.cracked = 0
        self.uncracked = 0


    def analyze_password(self, password):

        meets_policy, pass_length, charset, num_charsets, simplemask, advancedmask = self.policy.analyze_password(password)

        if meets_policy:
            self.complex_counter += 1
            stats_category = self.complex
        else:
            self.noncomplex_counter += 1
            stats_category = self.noncomplex

        self.overall['passwords'].increment(password)
        self.overall['basewords'].increment(password)
        self.overall['mutations'].increment(password)
        self.overall['advancedmasks'].increment(advancedmask)
        stats_category['length'].increment(pass_length)
        stats_category['entropy'].increment(password)
        stats_category['simplemasks'].increment(simplemask)
        stats_category['charactersets'].increment(charset)


    @property
    def meta(self):

        return PasswordMetaStat(
            self.cracked,
            self.uncracked,
            self.complex_counter,
            self.noncomplex_counter,
            self.complex['length'],
            self.noncomplex['length'],
            options=self.options
        )
    

    def analyze(self, iterator):
        ''' Generate password statistics. '''

        for line in iterator:
            if self.options.delimiter:
                try:
                    password = line.split(self.options.delimiter)[self.options.field-1]
                except IndexError:
                    self.uncracked += 1
                    continue
            else:
                password = line

            if len(password) > 0:
                self.cracked += 1
                self.analyze_password(password)
            else:
                self.uncracked += 1


    def __getattr__(self, attr):

        return self.complex[attr]


    def __iter__(self):

        for k,stat in self.overall.items():
            yield k,stat
        for k,stat in self.complex.items():
            yield k,stat
        for k,stat in self.noncomplex.items():
            yield k,stat


#!/usr/bin/env python3

# by TheTechromancer

import os
import re
import sys
import argparse
from password_stretcher.lib.utils import *
from password_stretcher.lib.errors import *
from password_stretcher.lib.mangler import *
from password_stretcher.lib.spider import Spider
from password_stretcher.lib.policy import PasswordPolicy



def stretcher(options):

    if options.minlength is not None and options.maxlength is not None:
        if options.minlength > options.maxlength:
            print('U WOT M8')
            sys.exit()

    show_written_count = not sys.stdout.isatty()
    written_count = 0

    policy = PasswordPolicy(
        minlength=options.minlength,
        maxlength=options.maxlength,
        mincharsets=options.mincharsets,
        required_charsets=options.charsets,
        regex=options.regex,
    )

    sys.stderr.write('[+] Reading input wordlist...')
    mangler = Mangler(
        _input=options.input,
        output_size=options.limit,
        double=options.double,
        perm=options.permutations,
        leet=options.leet,
        cap=options.cap,
        capswap=options.capswap,
        pend=options.pend,
    )
    sys.stderr.write(f' read {len(mangler.input):,} words {"(after basic cap mutations)" if (options.cap and not options.capswap) else ""}\n')
    if options.permutations > 1:
        sys.stderr.write(f'[*] Input wordlist after permutations: {len(mangler.mutators[0]):,}\n')
    else:
        sys.stderr.write(f'[*] Output capped at {mangler.output_size:,} words\n')
    if any([mangler.leet, mangler.cap, mangler.pend]):
        sys.stderr.write('[+] Mutations allowed per word:\n')
        for mutator in mangler.mutators[1:]:
            sys.stderr.write(f'       {str(mutator):<16}{mutator.limit:,}\n')
    if policy:
        sys.stderr.write(f'[+] Filtering based on policy, output size may be reduced\n')

    #sys.stderr.write(f'[+] Estimated output: {len(mangler):,} words\n')

    bytes_written = 0
    for mangled_word in mangler:

        if policy.meets_policy(mangled_word):
            sys.stdout.buffer.write(mangled_word + b'\n')
            bytes_written += (len(mangled_word)+1)
            if show_written_count and written_count % 10000 == 0:
                sys.stderr.write(f'\r[+] {written_count:,} words written ({bytes_to_human(bytes_written)})    ')

            written_count += 1

        else:
            # if the word didn't meet length requirements, increase the limit by 1
            mangler.mutators[-1].cur_limit += 1

    if show_written_count:
        sys.stderr.write(f'\r[+] {written_count:,} words written ({bytes_to_human(bytes_written)})    \n')

    sys.stdout.buffer.flush()
    sys.stdout.close()


def main():

    parser = argparse.ArgumentParser(description='FETCH THE PASSWORD STRETCHER')
    parser.add_argument('-i', '--input', nargs='+', default=ReadSTDIN(), help='input website or wordlist(s) (default: STDIN)', metavar='')
    parser.add_argument('--limit', type=human_to_int, help='limit length of output (default: max(100M, 1000x input))')
    mangling = argparse.ArgumentParser.add_argument_group(parser, 'mangling options')
    mangling.add_argument('-L', '--leet', action='store_true', help='"leetspeak" mutations')
    mangling.add_argument('-c', '--cap', action='store_true', help='common upper/lowercase variations')
    mangling.add_argument('-C', '--capswap', action='store_true', help='all possible case combinations')
    mangling.add_argument('-p', '--pend', action='store_true', help='append/prepend common digits & special characters')
    mangling.add_argument('-dd', '--double', action='store_true', help='double each word (e.g. "Pass" --> "PassPass")')
    mangling.add_argument('-P', '--permutations',type=int, default=1, help='max permutation depth (careful! massive output)', metavar='INT')
    filters = argparse.ArgumentParser.add_argument_group(parser, 'password complexity filters')
    filters.add_argument('--minlength', type=int, metavar='8', help='minimum password length')
    filters.add_argument('--maxlength', type=int, metavar='8', help='maximum password length')
    filters.add_argument('--mincharsets', type=int, metavar='3', help='must have this many character sets')
    filters.add_argument('--charsets', nargs='+', choices=PasswordPolicy.charset_choices, help='must include these character sets')
    filters.add_argument('--regex', type=re.compile, metavar='\'$[a-z]*^\'', help='custom regex')
    spidering = argparse.ArgumentParser.add_argument_group(parser, 'spider options')
    spidering.add_argument('--spider-depth', type=int, default=1, help='maximum website spider depth (default: 1)')
    spidering.add_argument('--user-agent', help='user-agent for web spider')

    try:

        options = parser.parse_args()

        # print help if there's nothing to stretch
        if type(options.input) == ReadSTDIN and stdin.isatty():
            parser.print_help()
            sys.stderr.write('\n\n[!] Please specify wordlist(s) or pipe to STDIN\n')
            exit(2)

        if not type(options.input) == ReadSTDIN:
            options.input = read_uris(options.input)

        if type(options.input) == Spider:
            spider = options.input
            spider.depth = options.spider_depth
            spider.user_agent = options.user_agent
            spider.start()

        stretcher(options)


    except re.error:
        print(f'[!] Invalid regex')
        print('[*] Remember to place regex in single quotes: \'^Password[0-9]+\'')
        sys.exit(2)
    except BrokenPipeError:
        # Python flushes standard streams on exit; redirect remaining output
        # to devnull to avoid another BrokenPipeError at shutdown
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        sys.exit(1)  # Python exits with error code 1 on EPIPE
    except (PasswordStretcherError, AssertionError) as e:
        sys.stderr.write(f'\n[!] {e}\n')
        exit(1)
    except argparse.ArgumentError:
        sys.stderr.write('\n[!] Check your syntax. Use -h for help.\n')
        exit(2)
    except KeyboardInterrupt:
        sys.stderr.write('\n[!] Interrupted.\n')
        exit(2)


if __name__ == '__main__':
    main()


#!/usr/bin/env python3

# by TheTechromancer

import re
import requests
import urllib.parse
from sys import stderr
from .errors import SpiderError
from .utils import url_to_domain
from html.parser import HTMLParser



class Parser(HTMLParser):
    '''
    Parses HTML data and extracts words
    Keeps count of each word in self.words
    '''

    def __init__(self):

        self.word_regex = re.compile(r'[a-z][a-z01357$@]+[a-z]', re.I)

        # track occurrences of each word
        self.words = dict()
        # stores each responses' links to other pages
        self.temp_links = set()

        super().__init__()


    def injest(self, html):

        self.temp_links.clear()
        self.feed(html)
        self.close()
        self.handle_words(html)
        return list(self.temp_links)


    def handle_words(self, html):

        # strip start tags
        #html = re.sub(r'<\w+', '', html)
        # strip end tags
        #html = re.sub(r'</\w+>', '', html)

        for word in self.word_regex.findall(html):
            try:
                self.words[word] += 1
            except KeyError:
                self.words[word] = 1

    '''
    def unknown_decl(self, data):

        self.handle_words(data)


    def handle_data(self, data):
        
        self.handle_words(data)


    def handle_comments(self, data):

        self.handle_words(data)
    '''

    def handle_starttag(self, tag, attrs):

        if tag == 'a':
            for attr, value in attrs:
                if attr == 'href' and value:
                    self.temp_links.add(value)




class Spider:

    def __init__(self, url, depth=2, user_agent=None):

        if user_agent is None:
            user_agent = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/89.0.4389.114 Safari/537.36 Edg/89.0.774.75'
        self.user_agent = user_agent

        self.url = url
        self.base_domain = url_to_domain(url)
        self.session = requests.Session()
        self.parser = Parser()
        self.visited = set()
        self.depth = depth
        self.headers = {
            'Upgrade-Insecure-Requests': '1',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9',
            'Sec-Fetch-Site': 'none',
            'Sec-Fetch-Mode': 'navigate',
            'Sec-Fetch-User': '?1',
            'Sec-Fetch-Dest': 'document',
            'Accept-Encoding': 'gzip, deflate, br',
            'Accept-Language': 'en-US,en;q=0.9'
        }


    def start(self):

        self.headers.update({'User-Agent': self.user_agent})

        try:
            self.get(self.url)
            stderr.write('\n')
            stderr.flush()
        except requests.RequestException:
            raise SpiderError(f'Error visiting URL: "{self.url}"')
        except KeyboardInterrupt:
            stderr.write('\n\n[!] Stopping spider...\n')


    def get(self, url, depth=None):

        if depth is None:
            depth = self.depth

        if not url in self.visited and depth > 0:

            response = self.session.get(url, headers=self.headers)
            self.visited.add(url)
            links = self.parser.injest(response.text)

            for link in links:
                link = urllib.parse.urljoin(url, link)
                try:
                    if url_to_domain(link) == self.base_domain:
                        self.get(link, depth=depth-1)
                except ValueError:
                    continue

        stderr.write(f'\r[+] Found {len(self.parser.words):,} words in {len(self.visited):,} pages')



    def __iter__(self):

        words = list(self.parser.words.items())
        words.sort(key=lambda x: x[1], reverse=True)

        for word, count in words:
            yield word.encode('utf-8')


#!/usr/bin/env python3

# by TheTechromancer

import os
from pathlib import Path
from .mutator import Mutator


class Pend(Mutator):

    scale = 5
    fname = 'append/prepend'

    def __init__(self, _input, limit=2048):

        self.rules = []
        self.read_rules()

        super().__init__(_input, limit)


    def __len__(self):

        return min(self.limit, len(self.rules))


    def mutate(self, word):

        yield word

        for rule in self.rules:
            yield rule.replace(b'\x00', word)


    def read_rules(self, rule_dir=None):

        if rule_dir is None:
            rule_dir = Path(__file__).resolve().parent.parent / 'lists'

        for _, _, files in os.walk(rule_dir):
            for file in files:
                if any(file.lower().endswith(x) for x in ['rule', 'rules']):
                    with open(rule_dir / file) as f:
                        lines = [l.strip('\r\n') for l in f.readlines()]
                        for line in lines:
                            try:
                                rule = self.parse_rule(line)
                                if len(rule) > 1:
                                    self.rules.append(rule)
                            except ValueError:
                                continue


    @staticmethod
    def parse_rule(rule):

        parsed_rule = []

        rule_chars = rule.split(' ')
        for char in rule_chars:
            if char.startswith('^'):
                if len(char) == 1:
                    parsed_rule.append(b' ')
                elif len(char) > 1:
                    parsed_rule.append(char[-1].encode('utf-8'))
                else:
                    raise ValueError

        # the word goes here
        parsed_rule.append(b'\x00')

        for char in rule_chars:
            if char.startswith('$'):
                if len(char) == 1:
                    parsed_rule.append(b' ')
                elif len(char) > 1:
                    parsed_rule.append(char[-1].encode('utf-8'))
                else:
                    raise ValueError

        return b''.join(parsed_rule)




#!/usr/bin/env python3

# by TheTechromancer

class PasswordStretcherError(Exception):
    pass

class SpiderError(PasswordStretcherError):
    pass

class InputListError(PasswordStretcherError):
    pass

class PasswordAnalyzerError(PasswordStretcherError):
    pass


#!/usr/bin/env python3

# by TheTechromancer

from .mutator import Mutator


class Cap(Mutator):

    # roughly the number of capital mutations when capswap is disabled
    cap_multiplier = 4
    scale = 2
    fname = 'capitalization'

    def __init__(self, _input, limit=256, capswap=False):

        self.capswap = capswap
        # the average number of words produced by the cap() (not capswap)

        super().__init__(_input, limit)


    def __len__(self):

        if not self.capswap:
            return self.cap_multiplier
        else:
            return self.limit


    def mutate(self, word):

        # always yield the most likely candidates first
        results = []
        for r in [word, word.lower(), word.upper(), word.swapcase(), word.capitalize(), word.title()]:
            if r not in results:
                results.append(r)
                yield r

        # then move on to full cap mutations if requested
        if self.capswap:
            for r in self._capswap(word):
                if not r in results:
                    yield r


    def _capswap(self, word):

         # many recursions make light work
        if len(word) == 1:
            yield word
            if word.isalpha():
                yield word.swapcase()

        else:
            mid_point = int(len(word)/2)
            for right_half in self._capswap(word[mid_point:]):
                for left_half in self._capswap(word[:mid_point]):
                    yield left_half + right_half



#!/usr/bin/env python3

# by TheTechromancer

from .cap import Cap
from .leet import Leet
from .pend import Pend
from .perm import Perm
from functools import reduce

class Mangler():

    def __init__(self, _input, output_size=None, double=False, perm=0, leet=False, cap=False, capswap=False, pend=False, key=lambda x: x):

        # load input list into memory and deduplicate
        if cap and not capswap:
            self.input = set(Cap(_input))
        else:
            self.input = set(_input)

        self.input = list(self.input)
        self.input.sort(key=lambda x: len(x))

        self.perm_depth = perm
        self.leet       = leet
        self.capswap    = capswap
        self.cap        = cap or capswap
        self.double     = double
        self.pend       = pend

        self.mutators = [Perm(self.input, double=double, perm_depth=perm)]

        if self.leet:
            self.mutators.append(Leet(self.mutators[-1]))
        if self.capswap:
            self.mutators.append(Cap(self.mutators[-1], capswap=True))
        if self.pend:
            self.mutators.append(Pend(self.mutators[-1]))

        if output_size:
            self.set_output_size(output_size)
        else:
            self.set_output_size(max(len(self.input)*1000, 100000000))


    def __iter__(self):
        '''
        generator function
        runs mangling functions on wordlist
        yields each mutated word
        '''

        for word in self.mutators[-1]:
            yield word


    def __len__(self):
        '''
        Estimates the total output length based on requested mangling parameters
        '''

        for m in self.mutators:
            length *= len(m)

        return int(length)


    def chain_mutators(self, _input=None, mutators=None):

        if mutators is None:
            mutators = self.mutators
        if _input is None:
            _input = self.perm()

        if len(mutators) == 0:
            return _input

        return self.chain_mutators(mutators[0].gen(_input), mutators[1:])



    def set_output_size(self, target_size):
        '''
        sets self.max_cap and self.max_leet based on desired output size

        (s1*m) * (s2*m) * (s3*m) = t
        where each () is a mutator
        t = targe size (known)
        m = multiplier (unknown)
        s = mutator's scale (known)
        solving for m
        '''

        self.output_size = target_size

        x = target_size / len(self.mutators[0])
        num_mutators = len(self.mutators[1:])

        # multiply scales together
        try:
            cumulative_scale = reduce(lambda x, y: x*y, [m.scale for m in self.mutators[1:]])
        except TypeError:
            cumulative_scale = 1

        try:
            x = x / cumulative_scale
        except ArithmeticError:
            pass

        try:
            x **= (1 / num_mutators)
            for mutator in self.mutators[1:]:
                mutator.limit = max(1, int(x * mutator.scale))
        except ArithmeticError:
            pass

        '''
        if self.capswap and self.leet:
            multiplier = (multiplier / 2) ** (1 / 2)
            self.max_cap = max(1, int(multiplier * 2))
            self.max_leet = max(1, int(multiplier))

        elif self.cap and self.leet:
            self.max_leet = int(multiplier / Cap.cap_multiplier)

        elif self.capswap:
            self.max_cap = max(1, int(multiplier))

        elif self.leet:
            self.max_leet = max(1, int(multiplier))
        '''


# by TheTechromancer

# This was adapted from the Password Statistical Analysis tool by Peter Kacherginsky
# https://github.com/iphelix/pack

class PasswordPolicy:

    charset_choices = ['numeric', 'loweralpha', 'upperalpha', 'special']

    charset_flags = {
        0b0001: 'numeric',
        0b0010: 'loweralpha',
        0b0100: 'upperalpha',
        0b1000: 'special',
        0b0110: 'mixedalpha',
        0b0011: 'loweralphanum',
        0b0101: 'upperalphanum',
        0b1010: 'loweralphaspecial',
        0b1100: 'upperalphaspecial',
        0b1001: 'specialnum',
        0b1110: 'mixedalphaspecial',
        0b1101: 'upperalphaspecialnum',
        0b1011: 'loweralphaspecialnum',
        0b0111: 'mixedalphanum',
        0b1111: 'all',
        'numeric': 0b0001,
        'loweralpha': 0b0010,
        'upperalpha': 0b0100,
        'special': 0b1000
    }

    def __init__(self, minlength=None, maxlength=None, mincharsets=None, required_charsets=None, regex=None):

        self.minlength = minlength
        self.maxlength = maxlength
        self.mincharsets = mincharsets
        self.regex = regex
        self.required_charset = None
        if required_charsets is not None:
            self.required_charset = 0b0000
            for charset in [c.strip().lower() for c in required_charsets]:
                if charset in self.charset_flags:
                    self.required_charset |= self.charset_flags[charset]


    def meets_policy(self, password, pass_length=None, charset=None):

        if type(password) == bytes:
            try:
                password = password.decode()
            except UnicodeError:
                password = str(password)[2:-1]

        #if pass_length is None or charset is None:
        meets_policy, pass_length, charset, num_charsets, simplemask, advancedmask = self.analyze_password(password, calc_policy=False)

        meets_policy = False
        if (self.maxlength is None          or pass_length <= self.maxlength) and \
           (self.minlength is None          or pass_length >= self.minlength) and \
           (self.regex is None              or self.regex.match(password)) and \
           (self.mincharsets is None        or num_charsets >= self.mincharsets) and \
           (self.required_charset is None   or self.required_charset & charset == self.required_charset):
           meets_policy = True
        return meets_policy


    def analyze_password(self, password, calc_policy=True):

        # Character-set flags
        charset = 0b0000
        required_charset = 0b0000
        simplemask = []
        advancedmask_string = []

        # Detect simple and advanced masks
        for letter in password:
 
            if letter.islower():
                charset |= 0b0010
                if calc_policy:
                    advancedmask_string.append('?l')
                    if not simplemask or not simplemask[-1] == 'Word': simplemask.append('Word')

            elif letter.isdigit():
                charset |= 0b0001
                if calc_policy:
                    advancedmask_string.append('?d')
                    if not simplemask or not simplemask[-1] == 'Number': simplemask.append('Number')

            elif letter.isupper():
                charset |= 0b0100
                if calc_policy:
                    advancedmask_string.append('?u')
                    if not simplemask or not simplemask[-1] == 'Word': simplemask.append('Word')

            else:
                charset |= 0b1000
                if calc_policy:
                    advancedmask_string.append('?s')
                    if not simplemask or not simplemask[-1] == 'Symbol': simplemask.append('Symbol')

        num_charsets = bin(charset).count('1')
        pass_length = len(password)
        meets_policy = None
        if calc_policy:
            meets_policy = self.meets_policy(password, pass_length, charset)

        return (
            meets_policy,
            pass_length,
            charset,
            num_charsets,
            '+'.join(simplemask) if len(simplemask) <= 3 else 'Other',
            ''.join(advancedmask_string)
        )


    def wrap(self, iterable):

        if not self:
            yield from iterable

        else:
            for password in iterable:
                if self.meets_policy(password):
                    yield password


    def __bool__(self):

        return not all([_ is None for _ in [self.minlength, self.maxlength, self.mincharsets, self.required_charset, self.regex]])


#!/usr/bin/env python3

# by TheTechromancer


class Mutator():
    '''
    base class for mutators like capswap, leet, and *pend
    '''

    # prioritization compared to other mutators when limit is being enforced
    # e.g. "pend" gets a scale of 5 while "leet" gets 1
    # this means that for every 1 leet mutation, there are 5 pend mutations
    scale = 1
    # friendly name to describe mutator type
    fname = 'mutator'

    def __init__(self, _input, limit=128):

        self.input = _input

        # maximum mutations per word
        self.limit = limit
        # carry over unused mutations into the next word
        self.cur_limit = 0


    def __len__(self):

        return self.limit


    def __iter__(self):

        for word in self.input:
            self.cur_limit += self.limit
            for r in self.mutate(word):
                if self.cur_limit > 0:
                    yield r
                    self.cur_limit -= 1
                else:
                    break


    def __str__(self):

        return self.fname


    def mutate(self, word):
        '''
        override in child class
        '''
        yield word


#!/usr/bin/env python3

# by TheTechromancer

import string
from sys import stdin
from pathlib import Path
from urllib.parse import urlparse
from .errors import InputListError


def read_uris(uris):

    files = []
    for uri in uris:
        if any([uri.startswith(x) for x in ['http://', 'https://']]):
            from .spider import Spider
            return Spider(uri)
        else:
            files.append(uri)

    return ReadFiles(*files)


class ReadFiles():

    def __init__(self, *filenames, binary=True):

        self.files = [ReadFile(filename, binary=binary) for filename in filenames]

    def __iter__(self):

        for file in self.files:
            yield from file


class ReadFile():

    def __init__(self, filename, binary=True):

        self.filename = Path(filename)
        if binary:
            self.mode = 'rb'
            self.strip = b'\r\n'
        else:
            self.mode = 'r'
            self.strip = '\r\n'

        if not self.filename.exists() or self.filename.is_dir():
            raise InputListError(f'Cannot find the file {self.filename}')



    def __iter__(self):

        with open(self.filename, self.mode) as f:
            i = f.__iter__()
            while 1:
                try:
                    line = next(i)
                    line = line.strip(self.strip)
                    if line:
                        yield line
                except StopIteration:
                    break
                except UnicodeError:
                    continue



class ReadSTDIN():

    def __init__(self, binary=True):

        if binary:
            self.buffer = stdin.buffer
            self.strip = b'\r\n'
        else:
            self.buffer = stdin
            self.strip = '\r\n'

    def __iter__(self):

        while 1:
            line = self.buffer.readline()
            if line:
                line = line.strip(self.strip)
                if line:
                    yield line
            else:
                break



def int_to_human(i):
    '''
    shortens large integer to human-readable format
    e.g. 1000 --> 1K
    '''

    sizes = ['', 'K', 'M', 'B', 'T']
    units = {}
    count = 0
    for size in sizes:
        units[size] = pow(1000, count)
        count +=1

    for size in sizes:
        if abs(i) < 1000.0:
            if size == sizes[0]:
                i = str(int(i))
            else:
                i = '{:.2f}'.format(i)
            return '{}{}'.format(i, size)
        i /= 1000

    raise ValueError


def human_to_int(h):
    '''
    converts human-readable number to integer
    e.g. 1K --> 1000
    '''

    if type(h) == int:
        return h

    units = {'': 1, 'K': 1000, 'M': 1000**2, 'B': 1000**3, 'T': 1000**4}

    try:
        h = h.upper().strip()
        i = float(''.join(c for c in h if c in string.digits + '.'))
        unit = ''.join([c for c in h if c in string.ascii_uppercase])
    except (ValueError, KeyError):
        raise ValueError(f'Invalid number "{h}"')

    return int(i * units[unit])


def bytes_to_human(_bytes):
    '''
    converts bytes to human-readable filesize
    e.g. 1024 --> 1KB
    '''

    sizes = ['B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB']
    units = {}
    for count,size in enumerate(sizes):
        units[size] = pow(1024, count)

    for size in sizes:
        if abs(_bytes) < 1024.0:
            if size == sizes[0]:
                _bytes = str(int(_bytes))
            else:
                _bytes = '{:.2f}'.format(_bytes)
            return '{}{}'.format(_bytes, size)
        _bytes /= 1024

    raise ValueError



def hostname_to_domain(hostname):

    hostname = hostname.lower().split('.')
    try:
        if hostname[-2] == 'co':
            return '.'.join(hostname[-3:])
        else:
            return '.'.join(hostname[-2:])
    except IndexError:
        return '.'.join(hostname)



def url_to_domain(url):

    try:
        return hostname_to_domain(urlparse(url).hostname).lower()
    except AttributeError:
        raise ValueError(f'Invalid URL: {url}')


def thread_wrapper(target, *args, **kwargs):

    try:
        target(*args, **kwargs)
    except KeyboardInterrupt:
        pass
    except Exception as e:
        import traceback
        traceback.print_exc()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['password_stretcher', 'password_stretcher.lib', 'password_stretcher.lib.stat']

package_data = \
{'': ['*'],
 'password_stretcher': ['lists/*'],
 'password_stretcher.lib.stat': ['assets/*']}

install_requires = \
['requests>=2.25.1,<3.0.0']

entry_points = \
{'console_scripts': ['password-stretcher = password_stretcher.stretcher:main']}

setup_kwargs = {
    'name': 'password-stretcher',
    'version': '1.0.1',
    'description': 'A deadly password mangler',
    'long_description': None,
    'author': 'TheTechromancer',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
